import logging
import sqlite3


LOG_SQL_TO_FILE = 0


log = logging.getLogger(__name__)

dbtype = None
db = None
param_mark = None

sqllog = None


def init(_dbtype, dbname):
    global dbtype, param_mark
    dbtype = _dbtype

    if dbtype == "pgsql":
        param_mark = "%s"
        init_pgsql(dbname)
    else:
        param_mark = "?"
        init_sqlite(dbname)


def init_sqlite(dbname):
    global db
    db = sqlite3.connect(dbname)
    db.row_factory = sqlite3.Row
    if LOG_SQL_TO_FILE:
        global sqllog
        sqllog = open(dbname + ".sql", "w")


def init_pgsql(dbname):
    import psycopg
    global db
    db = psycopg.connect(dbname, row_factory=psycopg.rows.namedtuple_row)
    execute_dml("SET session_replication_role = 'replica'")
    execute_dml("BEGIN")


def execute_dml(sql, values = (), returning=None):
    if returning:
        sql += " RETURNING " + returning
    if dbtype == "pgsql":
        sql = sql.replace("?", "%s")
    if LOG_SQL_TO_FILE:
        sqllog.write("%s %s\n" % (sql, values))
        return
    cursor = db.cursor()
    log.debug(sql + " " + str(values))
    cursor.execute(sql, values)
    if dbtype == "pgsql":
        if returning:
            return cursor.fetchone()[0]
    else:
        return cursor.lastrowid


def executescript(sql):
    cursor = db.cursor()
    log.debug(sql)
    if dbtype == "pgsql":
        cursor.execute(sql)
    else:
        cursor.executescript(sql)


def insert(table, fields=None, or_replace=False, returning=None, **kw):
    repl_clause = ""
    if or_replace:
        repl_clause = " OR REPLACE"
    if fields is None:
        fields = kw
    field_names = []
    field_vals = []
    qmarks = []
    for k, v in fields.items():
        field_names.append(k)
        field_vals.append(v)
        qmarks.append(param_mark)
    sql = "INSERT%s INTO %s(%s) VALUES (%s)" % (repl_clause, table, ", ".join(field_names), ", ".join(qmarks))
    id = execute_dml(sql, field_vals, returning)
    return id


def select(table, where=None, order=None):
    cursor = db.cursor()
    if where is None:
        where = ""
    else:
        where = " WHERE " + where
    if order is None:
        order = ""
    else:
        order = " ORDER BY " + order
    sql = "SELECT * FROM %s%s%s" % (table, where, order)
    if LOG_SQL_TO_FILE:
        sqllog.write("%s\n" % sql)
    cursor.execute(sql)
    log.debug(sql)
    return cursor.fetchall()


def commit():
    log.debug("COMMIT")
    db.commit()



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging

import pandera.pandas as pa
from pandera.errors import SchemaError
from pandera.typing import DataFrame

from .schemas import AccountType, TransactionSchema, AccountSchema, SecuritySchema, SecurityPriceSchema, Account, \
    Security
from ..exceptions import InputError

log = logging.getLogger(__name__)


class Portfolio:
    _accounts: DataFrame[AccountSchema] = AccountSchema.empty()
    _securities: DataFrame[SecuritySchema] = SecuritySchema.empty()
    _transactions: DataFrame[TransactionSchema] = TransactionSchema.empty()
    _prices: DataFrame[SecurityPriceSchema] = SecurityPriceSchema.empty()
    _attributes: dict[str, dict[str, str]] = {}
    base_currency: str = ''

    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            self,
            accounts: DataFrame[AccountSchema] | None = None,
            transactions: DataFrame[TransactionSchema] | None = None,
            securities: DataFrame[SecuritySchema] | None = None,
            prices: DataFrame[SecurityPriceSchema] | None = None,
            attributes: dict[str, dict[str, str]] | None = None
    ):
        if accounts is not None:
            try:
                self._accounts = AccountSchema.validate(accounts)
            except SchemaError as e:
                log.error('accounts schema invalid: %s', e)

        if securities is not None:
            try:
                self._securities = SecuritySchema.validate(securities)
            except SchemaError as e:
                log.error('securities schema invalid: %s', e)

        if transactions is not None:
            try:
                self._transactions = TransactionSchema.validate(transactions)
            except SchemaError as e:
                log.error('transactions schema invalid: %s', e)

        if prices is not None:
            try:
                self._prices = SecurityPriceSchema.validate(prices)
            except SchemaError as e:
                log.error('security prices schema invalid: %s', e)

        self._attributes = attributes if attributes is not None else {}

    @property
    def securities_accounts(self) -> DataFrame[AccountSchema]:
        return AccountSchema.validate(self._accounts[self._accounts['type'] == AccountType.SECURITIES.value])

    @property
    def deposit_accounts(self) -> DataFrame[AccountSchema]:
        return AccountSchema.validate(self._accounts[self._accounts['type'] == AccountType.DEPOSIT.value])

    @property
    def securities_account_transactions(self) -> DataFrame[TransactionSchema]:
        return TransactionSchema.validate(self._transactions[self._transactions['accountType'] == AccountType.SECURITIES.value].sort_values(by=['date']))

    @property
    def deposit_account_transactions(self) -> DataFrame[TransactionSchema]:
        return TransactionSchema.validate(self._transactions[self._transactions['accountType'] == AccountType.DEPOSIT.value].sort_values(by=['date']))

    @property
    @pa.check_types
    def securities(self) -> DataFrame[SecuritySchema]:
        return self._securities

    @property
    @pa.check_types
    def prices(self) -> DataFrame[SecurityPriceSchema]:
        return self._prices

    @property
    def all_attributes(self) -> dict[str, str]:
        return {uuid: name for attributes in self._attributes.values() for uuid,name in attributes.items()}

    @property
    def security_attributes(self) -> dict[str, str]:
        return self._attributes.get('securities', {})

    @property
    def account_attributes(self) -> dict[str, str]:
        return self._attributes.get('accounts', {})


def get_securities_account_by_id(portfolio: Portfolio, account_id: str) -> Account:
    if account_id not in portfolio.securities_accounts.index:
        raise InputError(f"Securities account '{account_id}' not found in portfolio")

    return Account(**portfolio.securities_accounts.reset_index().set_index('accountId', drop=False).loc[account_id])


def get_security_by_id(portfolio: Portfolio, security_id: str) -> Security:
    if security_id not in portfolio.securities.index:
        raise InputError(f"Security '{security_id}' not found in portfolio")

    return Security(**portfolio.securities.reset_index().set_index('securityId', drop=False).loc[security_id])



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
import shutil
from pathlib import Path
from typing import cast

from rich.console import Console
import typer

from pp_terminal.exceptions import InputError
from pp_terminal.output.strategy import OutputStrategy

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)


@app.command(name="export")
def export_file(
    ctx: typer.Context,
    output_file: Path = typer.Option(
        ...,
        help="Output path for XML file",
        file_okay=True,
        dir_okay=False,
    ),
) -> None:
    """
    Export Portfolio Performance XML file.

    When used with --anonymize flag, exports the anonymized version.
    Otherwise exports the original file.
    """

    source_file = cast(Path, ctx.obj.source_file)
    output = cast(OutputStrategy, ctx.obj.output)

    if output_file.exists():
        raise InputError(f"Output file {output_file} already exists")

    shutil.copy(source_file, output_file)
    console.print(output.text(f"Portfolio Performance file saved to {output_file}"))



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime
import logging

import pandas as pd
from pandera.typing import DataFrame

from pp_terminal.data.filters import filter_by_type
from pp_terminal.data.tax import calculate_prepaid_tax_per_lot
from pp_terminal.domain.schemas import TransactionType, Money, TransactionSchema, TaxPaidSchema, TaxLotSchema, Percent
from pp_terminal.exceptions import InputError

log = logging.getLogger(__name__)


def _filter_purchase_transactions(transactions: DataFrame[TransactionSchema]) -> DataFrame[TransactionSchema]:
    valid_purchases = transactions.pipe(filter_by_type, transaction_types=[TransactionType.BUY, TransactionType.DELIVERY_INBOUND]).sort_index(level='date')
    valid_purchases = valid_purchases[valid_purchases['shares'] > 0].copy()

    return TransactionSchema.validate(valid_purchases)


def _get_remaining_lots_after_fifo_matching(
    transactions: DataFrame[TransactionSchema]
) -> DataFrame[TaxLotSchema]:
    """
    Match all sell transactions to purchase lots using FIFO and return remaining lots.

    Returns:
        DataFrame of remaining lots after sales are matched.
        Each lot's 'shares' field represents remaining quantity.
        Exhausted lots (shares = 0) are removed.

    Implementation Note:
        FIFO matching is inherently sequential (each sale consumes lots in order,
        affecting state for the next sale). While the function accepts/returns DataFrames
        for schema validation and interface consistency, internally we use list of dicts
        for ~10x faster mutation during the matching algorithm. DataFrame .loc[] access
        has significant overhead that doesn't add value for sequential state updates.
    """
    lots = _filter_purchase_transactions(transactions)
    lots['purchasePrice'] = lots['amount'].abs() / lots['shares']  # save actual market price per share
    lots['fees'] = lots['fees'].fillna(0)
    lots = lots.rename(columns={'amount': 'costBasis'})

    if lots.empty:
        return TaxLotSchema.validate(lots)

    sell_transactions = transactions.pipe(filter_by_type, transaction_types=[TransactionType.SELL, TransactionType.DELIVERY_OUTBOUND])
    if sell_transactions.empty:
        return TaxLotSchema.validate(lots)

    # Convert to list of dicts for fast mutation during FIFO matching
    remaining_lots = lots.reset_index().to_dict('records')
    sales_sorted = sell_transactions.sort_index(level='date')

    for (_date, account_id, _security_id), row in sales_sorted.iterrows():
        shares_to_match = float(row['shares'])

        # Match against lots in FIFO order (only from same account)
        for lot in remaining_lots:
            if shares_to_match <= 0:
                break

            if lot['accountId'] != account_id:
                continue

            # Consume shares from this lot
            lot_shares = lot['shares']
            shares_from_lot = min(shares_to_match, lot_shares)
            new_shares = lot_shares - shares_from_lot
            shares_to_match -= shares_from_lot

            # Proportionally reduce fees based on remaining shares
            if lot_shares > 0:
                lot['fees'] = lot['fees'] * (new_shares / lot_shares)

            lot['shares'] = new_shares

        if shares_to_match > 0.0001:  # Allow small floating point errors
            log.warning('Sale of %.8f shares for security %s could not be fully matched to purchase lots', shares_to_match, _security_id)

    # Filter out exhausted lots and convert back to DataFrame
    remaining_lots = [lot for lot in remaining_lots if lot['shares'] > 0.0001]
    if not remaining_lots:
        return TaxLotSchema.empty()

    return TaxLotSchema.validate(pd.DataFrame(remaining_lots).set_index(['date', 'accountId', 'securityId']))


def _reduce_shares(remaining_lots_df: DataFrame[TaxLotSchema], shares_to_sell: float) -> DataFrame[TaxLotSchema]:
    # Calculate cumulative shares to determine lot consumption
    cumsum = remaining_lots_df['shares'].cumsum()
    prev_cumsum = cumsum.shift(1, fill_value=0.0)

    # Shares to take from each lot: min(lot_shares, remaining_needed)
    shares_taken = (shares_to_sell - prev_cumsum).clip(lower=0, upper=remaining_lots_df['shares'])

    # Filter to contributing lots only
    contributing_mask = shares_taken > 0
    if not contributing_mask.any():
        raise InputError(f"Insufficient shares available. Requested: {shares_to_sell}, Available: 0")

    df = remaining_lots_df[contributing_mask].copy()
    df['shares'] = shares_taken[contributing_mask].values

    # Validate sufficient shares
    total_allocated = df['shares'].sum()
    if total_allocated < shares_to_sell - 0.0001:  # Allow small floating point errors
        raise InputError(f"Insufficient shares available. Requested: {shares_to_sell}, Available: {total_allocated}")

    return TaxLotSchema.validate(df)


def _calculate_cost_basis(df: DataFrame[TaxLotSchema]) -> DataFrame[TaxLotSchema]:
    df = df.copy()
    df['costBasis'] = df['purchasePrice'] * df['shares'] + df['fees'].fillna(0)
    return TaxLotSchema.validate(df)


def _calculate_gains(df: DataFrame[TaxLotSchema]) -> DataFrame[TaxLotSchema]:
    df = _calculate_cost_basis(df)
    df['grossProceeds'] = df['shares'] * df['salePrice']
    df['capitalGain'] = df['grossProceeds'] - df['costBasis']
    df['taxableGain'] = df.apply(lambda row: max(0.0, row['capitalGain'] - row['prepaidTax']), axis=1)

    return TaxLotSchema.validate(df)


def calculate_fifo_sell(  # pylint: disable=too-many-locals,too-many-arguments,too-many-positional-arguments
        transactions: DataFrame[TransactionSchema],
        sell_date: datetime,
        sell_price: Money,
        tax_rate: Percent,
        shares_to_sell: float | None = None,
        tax_csv_data: DataFrame[TaxPaidSchema] | None = None
) -> DataFrame[TaxLotSchema]:
    """Calculate FIFO lots for shares being sold, including prepaid tax calculations."""
    df = transactions.copy()

    df = _get_remaining_lots_after_fifo_matching(df)
    if df.empty:
        return TaxLotSchema.empty()

    if shares_to_sell is not None:
        df = _reduce_shares(df, shares_to_sell)

    df = TaxLotSchema.validate(df)
    df['salePrice'] = sell_price

    df['prepaidTax'] = calculate_prepaid_tax_per_lot(df, sell_date, tax_csv_data).values

    df = _calculate_gains(df)
    df['totalTax'] = df['taxableGain'] * (tax_rate / 100.0)
    df['netProceeds'] = df['grossProceeds'] - df['totalTax']

    return TaxLotSchema.validate(df)


def calculate_total_cost_basis(transactions: DataFrame[TransactionSchema]) -> Money:
    """
    Calculate the cost basis of currently held shares for a security, i.e. what did I originally pay for the shares I currently hold?
    @link https://www.investopedia.com/terms/c/costbasis.asp
    """
    df = _get_remaining_lots_after_fifo_matching(transactions)
    df = _calculate_cost_basis(df)

    return Money(df['costBasis'].abs().sum())



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import re
from typing import Literal, Callable, Any

import pandas as pd
from rich.table import Table
from rich.text import Text

from pp_terminal.data.filters import drop_empty_values
from pp_terminal.utils.helper import format_money, format_shares
from pp_terminal.domain.schemas import Money


def camel_case_to_title(column_name: str) -> str:
    """
    Convert camelCase column names to Title Case with proper acronym handling.

    Examples:
        accountId -> Account ID
        securityId -> Security ID
        isRetired -> Is Retired
        wkn -> WKN
        isin -> ISIN
    """
    acronyms = {'Id': 'ID', 'Wkn': 'WKN', 'Isin': 'ISIN', 'Eur': 'EUR', 'Usd': 'USD'}

    # Insert space before uppercase letters that follow lowercase letters
    # accountId -> account Id
    text = re.sub('([a-z])([A-Z])', r'\1 \2', column_name)

    # Insert space before uppercase letter followed by lowercase (handles sequences of caps)
    # XMLParser -> XML Parser
    text = re.sub(r'(?<=[A-Z])(?=[A-Z][a-z])', ' ', text)

    # Title case each word
    text = text.title()

    # Replace known acronyms with uppercase versions
    for acronym, replacement in acronyms.items():
        text = re.sub(rf'\b{acronym}\b', replacement, text)

    return text


def format_value(value: Any, column_name: str, row: pd.Series) -> str:
    if column_name.lower() == 'shares' and isinstance(value, float):
        return format_shares(value)

    # Skip Purchase Price total (summing prices is meaningless)
    if column_name.lower().endswith('price') and row.isin(['Total']).any():
        return ''

    if isinstance(value, Money):
        return format_money(float(value), row['currency'] if 'currency' in row else column_name)

    return str(value)


class TableOptions:  # pylint: disable=too-few-public-methods
    def __init__(  # pylint: disable=too-many-arguments,too-many-positional-arguments
            self,
            title: str = '',
            caption: str = '',
            show_index: bool = True,
            show_total: bool = True,
            footer_lines: int = 0,
            value_formatter: Callable[[Any, str, pd.Series], str] = format_value
    ) -> None:
        self.title = title
        self.caption = caption
        self.show_index = show_index
        self.show_total = show_total
        self.footer_lines = footer_lines
        self.value_formatter = value_formatter


class TableDecorator(Table):
    _options: TableOptions

    def __init__(self, options: TableOptions) -> None:
        self._options = options

        super().__init__(show_footer=self.show_default_footer, title=options.title, caption=options.caption)

    @property
    def show_default_footer(self) -> bool:
        return self._options.show_total and self._options.footer_lines == 0  # multiple footer lines are not supported in rich by default

    def add_df(self, df: pd.DataFrame) -> Table:
        df = df.pipe(drop_empty_values)
        if df.empty:
            return self

        summary_row = (df.iloc[:-self._options.footer_lines] if self._options.footer_lines > 0 else df).select_dtypes(include='number').sum()  # only sum up numeric values
        if 'currency' in df.columns:
            summary_row['currency'] = df['currency'].mode()[0]
        if not self._options.show_index:
            # Find first non-currency column to put 'Total' label
            first_col = next((col for col in df.columns if col != 'currency'), None)
            if first_col:
                summary_row = pd.concat([summary_row, pd.Series(['Total'], index=[first_col])])

        # in case we have multiple footer lines, insert the total value into the right position in the dataframe
        footer_rows = None
        if self._options.show_total and not self.show_default_footer:
            footer_rows = df.iloc[-self._options.footer_lines:, :]
            df = df.iloc[:-self._options.footer_lines, :]
            footer_rows = pd.concat([df, summary_row.to_frame().T, footer_rows], ignore_index=True)

        # add index column if show_index is enabled
        if self._options.show_index:
            footer_value = 'Total' if self._options.show_total else ''
            self.add_column('ID', footer=footer_value if self.show_default_footer else '', justify='left')

        # add DataFrame columns to the table
        for i, column in enumerate(df.columns):
            if str(column) == 'currency':
                continue

            footer_value = self._options.value_formatter(summary_row[column], str(column), summary_row) if self._options.show_total and column in summary_row.index else ''
            justify = 'right' if column in summary_row.index and isinstance(summary_row[column], float) else 'left'  # type: Literal["right", "left"]

            if not self._options.show_index and footer_value == '' and i == 0:  # column is non-numeric
                footer_value = 'Total'

            column_title = camel_case_to_title(str(column))

            self.add_column(column_title, footer=footer_value if self.show_default_footer else '', justify=justify)

        for row_data in self._prepare_rows(df):
            self.add_row(*row_data)

        if footer_rows is None:
            return self

        self._print_footer(footer_rows)

        return self

    def _print_footer(self, df: pd.DataFrame) -> None:
        self.add_section()

        i = 0
        for row_data in self._prepare_rows(df.iloc[-(self._options.footer_lines + 1):]):
            self.add_row(*list(map(lambda x: Text(x, style='bold') if i == 0 else x, row_data)))  # pylint: disable=cell-var-from-loop
            i += 1

    def _prepare_rows(self, df: pd.DataFrame) -> list[list[str]]:
        rows = []
        for index, row in df.iterrows():
            row_data = [str(index)] if self._options.show_index else []
            row_data.extend([self._options.value_formatter(value, str(index), row) for index, value in row.drop('currency', errors='ignore').items()])
            rows.append(row_data)

        return rows



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime
import logging
from typing import Any, cast

import pandas as pd
import typer
from typing_extensions import Annotated
import numpy as np
from pandera.typing import DataFrame

from pp_terminal.data.filters import filter_by_type, drop_empty_values
from pp_terminal.utils.config import Config
from pp_terminal.utils.helper import get_last_year, footer
from pp_terminal.utils.options import tax_rate_callback, exemption_rate_callback
from pp_terminal.output.strategy import OutputStrategy, Console
from pp_terminal.output.strategy_excel import ExcelOutputStrategy
from pp_terminal.domain.portfolio_snapshot import PortfolioSnapshot, _NEGATIVE_SECURITIES_ACCOUNT_TRANSACTION_TYPES
from pp_terminal.domain.portfolio import Portfolio
from pp_terminal.domain.schemas import TransactionType, Percent, Money, VapResultSchema
from pp_terminal.output.table_decorator import TableOptions, format_value

app = typer.Typer()
console = Console()
log = logging.getLogger(__name__)

_debug_data: dict[str, pd.DataFrame | None] = {}
begin = None  # pylint: disable=invalid-name

# Basiszinssatz (base interest rate) by year for German tax calculations
# @link https://www.bundesbank.de/de/statistiken/geld-und-kapitalmaerkte/zinssaetze-und-renditen/basiszinssatz
_BASE_RATE_BY_YEAR: dict[int, Percent] = {
    2016: 1.1,
    2018: 0.87,
    2019: 0.52,
    2020: 0.07,
    2021: -0.45,
    2022: -0.05,
    2023: 2.55,
    2024: 2.29,
    2025: 2.53,
    2026: 3.2,
}


# @see https://www.gesetze-im-internet.de/invstg_2018/__18.html
def calculate(  # pylint: disable=too-many-locals,too-many-arguments,too-many-positional-arguments
        snapshot_period_begin: PortfolioSnapshot,
        snapshot_period_end: PortfolioSnapshot,
        base_rate_percent: Percent,
        tax_rate_percent: Percent,
        default_exemption_rate_percent: Percent = 30.0,
        exempt_rate_attr_uuid: str | None = None
) -> DataFrame[VapResultSchema] | None:
    base_rate = max(base_rate_percent, 0) / 100

    payouts = _calculate_payouts(snapshot_period_end)
    _debug_data['payouts'] = payouts.reset_index() if payouts is not None else None

    # @todo convert all values to EUR with rates from ECB, for the moment we simply remove currency
    # Calculate begin values only for shares held continuously from year start to year end
    shares_begin = snapshot_period_begin.shares
    shares_end = snapshot_period_end.shares

    if shares_begin is not None and shares_end is not None and not shares_begin.empty and not shares_end.empty:
        # Calculate minimum position during the year to detect complete sells
        min_shares_during_year = _calculate_minimum_shares_during_year(snapshot_period_end)

        if min_shares_during_year is not None:
            # Only count shares held continuously (never went to zero)
            effective_begin_shares = shares_begin.combine(min_shares_during_year, min, fill_value=0).clip(lower=0)
        else:
            # Fallback: cap by end shares (assumes continuous holding)
            effective_begin_shares = shares_begin.combine(shares_end, min, fill_value=0).clip(lower=0)

        effective_begin_shares.name = 'shares'

        # Calculate begin values using continuously held shares
        begin_values_in_eur = (snapshot_period_begin.latest_prices * effective_begin_shares).groupby(['accountId', 'securityId']).sum()
    else:
        begin_values_in_eur = snapshot_period_begin.values.groupby(['accountId', 'securityId']).sum()

    end_values_in_eur = snapshot_period_end.values.groupby(['accountId', 'securityId']).sum()

    # use df.subtract to align both matrices
    outcome = end_values_in_eur.subtract(begin_values_in_eur, fill_value=0)
    outcome.name = 'Outcome'
    _debug_data['outcome'] = outcome.reset_index()

    # for securities that have been bought within the year we need to take the number of months held into account
    pro_rata_shares = _calculate_prorata_shares_for_inyear_buys(snapshot_period_end)
    modified_values_begin = begin_values_in_eur.add(pro_rata_shares.mul(snapshot_period_begin.latest_prices, fill_value=0), fill_value=0) if pro_rata_shares is not None else snapshot_period_begin.values

    base_yield = modified_values_begin * base_rate * 0.7
    base_yield = outcome.combine(base_yield, np.minimum)
    base_yield.name = 'Base Yield'
    _debug_data['base_yield'] = base_yield.reset_index()

    vap = base_yield.subtract(payouts, fill_value=0) if payouts is not None else base_yield
    vap = vap.clip(lower=0).fillna(0)  # replace negative values with zero

    vap = vap * tax_rate_percent / 100

    # Apply exemption rate if configured
    if exempt_rate_attr_uuid and exempt_rate_attr_uuid in snapshot_period_end.portfolio.securities.columns:
        exempt_rate_per_security = (1 - snapshot_period_end.portfolio.securities[[exempt_rate_attr_uuid]]
                                    .astype(float)
                                    .fillna(default_exemption_rate_percent / 100)
                                    .rename(columns={exempt_rate_attr_uuid: 0}))  # column name must match vap
        vap = exempt_rate_per_security.mul(vap.to_frame(), level='securityId')

    if not vap.empty:
        vap = vap.unstack(level='accountId')
        # Only extract from tuple if it's a MultiIndex
        if isinstance(vap.columns, pd.MultiIndex):
            vap.columns = [col[1] if len(col) > 1 else col[0] for col in vap.columns]

    vap = vap.pipe(drop_empty_values)
    if vap.empty:
        return None

    vap = pd.merge(snapshot_period_end.portfolio.securities[['wkn', 'name', 'currency']], vap, left_index=True, right_index=True, how='right', validate='one_to_one').sort_values(by='name')

    securities_accounts = snapshot_period_end.portfolio.securities_accounts
    if not securities_accounts.empty and 'referenceAccount' in securities_accounts:
        # add the reference account balance
        vap.loc[len(vap)] = (
            pd.merge(
                securities_accounts,
                snapshot_period_end.balances.groupby(['accountId']).sum(),
                left_on='referenceAccount',
                right_index=True,
                how='left',
                validate='many_to_one'
            )['balance'].dropna().to_dict()
            | {'name': 'Related Account Balance', 'currency': snapshot_period_end.portfolio.base_currency}
        )

    result = vap.rename(columns=securities_accounts['name'])
    return VapResultSchema.validate(result)


def _calculate_payouts(snapshot_end: PortfolioSnapshot) -> pd.Series:
    transactions = snapshot_end.securities_account_transactions
    transactions = transactions[transactions.index.get_level_values('date').year == snapshot_end.date.year] if not transactions.index.get_level_values('date').empty else transactions

    payouts = transactions.pipe(filter_by_type, transaction_types=TransactionType.DIVIDENDS).groupby(['accountId', 'securityId'])['amount'].sum()
    payouts.name = 'Payouts'

    return payouts


def _calculate_minimum_shares_during_year(snapshot_end: PortfolioSnapshot) -> pd.Series | None:  # pylint: disable=too-many-locals
    """
    Calculate the minimum share position during the tax year for each security.
    This detects if a position went to zero (complete sell) and was then rebought.
    Returns minimum shares held at any point during the year.
    """
    transactions = snapshot_end.securities_account_transactions

    # Get year-start position
    year_start = datetime(snapshot_end.date.year, 1, 2)
    snapshot_begin = PortfolioSnapshot(snapshot_end.portfolio, year_start)
    begin_shares = snapshot_begin.shares

    if begin_shares is None or begin_shares.empty:
        # No position at year start, minimum is 0 for all
        return None

    # Get only in-year transactions
    transactions_inyear = transactions[
        transactions.index.get_level_values('date').year == snapshot_end.date.year
    ] if not transactions.index.get_level_values('date').empty else pd.DataFrame()

    if transactions_inyear.empty:
        # No transactions during year, minimum = begin shares
        return begin_shares

    # Calculate cumulative changes during the year
    def sign_shares(row: pd.Series) -> float:
        return float(-row['shares'] if row['type'] in [t.value for t in _NEGATIVE_SECURITIES_ACCOUNT_TRANSACTION_TYPES] else row['shares'])

    # Group by account/security and calculate minimum cumulative position
    result_dict = {}

    for (account_id, security_id, currency), begin_count in begin_shares.items():
        # Get transactions for this specific account/security
        mask = (
            (transactions_inyear.index.get_level_values('accountId') == account_id) &
            (transactions_inyear.index.get_level_values('securityId') == security_id)
        )
        security_txns = transactions_inyear[mask].copy() if mask.any() else pd.DataFrame()

        if security_txns.empty:
            # No transactions, min = begin
            result_dict[(account_id, security_id, currency)] = begin_count
        else:
            # Calculate cumulative position starting from begin_count
            security_txns = security_txns.reset_index().sort_values('date')
            security_txns['signed_shares'] = security_txns.apply(sign_shares, axis=1)
            cumulative = begin_count + security_txns['signed_shares'].cumsum()
            min_position = min(begin_count, cumulative.min())
            result_dict[(account_id, security_id, currency)] = max(0, min_position)

    if not result_dict:
        return None

    index = pd.MultiIndex.from_tuples(result_dict.keys(), names=['accountId', 'securityId', 'currency'])
    return pd.Series(list(result_dict.values()), index=index, name='MinShares')


def _calculate_prorata_shares_for_inyear_buys(snapshot_end: PortfolioSnapshot) -> pd.Series | None:
    transactions = snapshot_end.securities_account_transactions

    transactions_inyear = transactions[transactions.index.get_level_values('date').year == snapshot_end.date.year] if not transactions.index.get_level_values('date').empty else None
    if transactions_inyear is None:
        return pd.Series([], name='Amount', index=pd.MultiIndex.from_tuples([], names=['accountId', 'securityId']), dtype='float64')

    transactions_inyear = transactions_inyear.pipe(filter_by_type, transaction_types=[TransactionType.BUY, TransactionType.DELIVERY_INBOUND])
    transactions_inyear['months_held'] = snapshot_end.date.month - transactions_inyear.index.get_level_values('date').month + 1
    transactions_inyear['shares_original'] = transactions_inyear['shares']
    transactions_inyear['shares'] = transactions_inyear['shares'] * transactions_inyear['months_held']/12
    _debug_data['transactions_inyear'] = transactions_inyear[['shares_original', 'months_held', 'shares']].reset_index(level='date', drop=True).sort_values(by=['accountId', 'securityId', 'months_held']).reset_index()

    return transactions_inyear.groupby(['accountId', 'securityId'])['shares'].sum().abs()


def set_begin(value: datetime | None) -> datetime | None:
    """
    Temporary store the non-empty year / datetime in a global state.
    This is necessary because typer.default_factory does not have context available to make one option dependent on the other.
    """
    global begin  # pylint: disable=global-statement

    if value is not None:
        begin = value

    return value


def get_base_rate_percent_by_year() -> Percent | None:
    """Get the base rate (Basiszinssatz) for the selected year."""
    if begin is None:
        return None

    return _BASE_RATE_BY_YEAR.get(begin.year, 3.2)


@app.command(name="vap")
def print_tax_table(  # pylint: disable=too-many-locals
        ctx: typer.Context,
        year: Annotated[datetime, typer.Option(formats=["%Y"], help="The year to calculate the preliminary tax for", prompt=True, callback=set_begin, default_factory=get_last_year)],
        base_rate: Annotated[Percent, typer.Option(help="The base rate (Basiszinssatz)", min=-100, max=100, prompt="Base Rate (%)", prompt_required=True, default_factory=get_base_rate_percent_by_year)],
        tax_rate: Annotated[Percent, typer.Option(help="Your personal tax rate", min=0, max=100, callback=tax_rate_callback)] = None,  # type: ignore
        exemption_rate: Annotated[Percent, typer.Option(help="Default exemption rate (Teilfreistellung), can be overwritten for each security.", min=0, max=100, callback=exemption_rate_callback)] = None  # type: ignore
) -> None:
    """
    Show a detailed table with calculated German preliminary tax values ("Vorabpauschale"/VAP) for a specified year, per each security and account.
    """
    _debug_data.clear()

    portfolio = cast(Portfolio, ctx.obj.portfolio)
    output = cast(OutputStrategy, ctx.obj.output)
    config = cast(Config, ctx.obj.config)

    exempt_rate_uuid = None
    if config:
        exempt_rate_uuid = config.get('tax', {}).get('exemption-rate-attribute')

    console.print(output.hint('You can define the exemption rate per each security individually by creating a custom security attribute of type "Percent Number" in Portfolio Performance and add it to pp-terminal configuration file.'))

    snapshot_begin = PortfolioSnapshot(portfolio, datetime(year.year, 1, 2))
    snapshot_end = PortfolioSnapshot(portfolio, datetime(year.year, 12, 31))

    result = calculate(snapshot_begin, snapshot_end, base_rate, tax_rate, exemption_rate, exempt_rate_uuid)
    result = result.round(2) if result is not None else result

    vap_totals = {}
    if result is not None and not result.empty:
        balance_row_index = result[result['name'] == 'Related Account Balance'].index
        if len(balance_row_index) > 0:
            vap_data = result.drop(balance_row_index)
            account_columns = [col for col in result.columns if col not in ['wkn', 'name', 'currency']]
            vap_totals = vap_data[account_columns].sum().to_dict()

    def format_value_with_balance_check(value: Any, index: str, row: pd.Series) -> str:
        if 'name' in row.index and row['name'] == 'Related Account Balance' and isinstance(value, Money) and index in vap_totals:
            color = 'red' if value < vap_totals[index] else 'green'
            return f"[{color}]{format_value(value, index, row)}[/{color}]"
        return format_value(value, index, row)

    console.print(*output.result_table(
        result,
        TableOptions(
            title=f"Estimated Taxes on Vorabpauschale {year.year} (§18 InvStG)",
            show_index=False,
            footer_lines=1,
            value_formatter=format_value_with_balance_check
        )
    ))

    # Add debug sheets if verbose and Excel format
    verbose = cast(bool, ctx.obj.verbose)
    is_excel = isinstance(output, ExcelOutputStrategy)

    for key, df in _debug_data.items():
        if df is None:
            continue

        if verbose and is_excel:
            title = '(DBG) ' + key.replace('_', ' ').title()
            result_output = output.result_table(
                df,
                TableOptions(
                    title=title,
                    show_index=False,
                    show_total=False
                )
            )
            if result_output is not None:
                console.print(*result_output)
        else:
            log.debug(df)

    console.print(output.warning('This simulation assumes that all amounts are in EUR excl. Sparerpauschbetrag.'))
    console.print(output.text(footer()), style="dim")



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from datetime import datetime

import pandas as pd
from pandera.typing import DataFrame

from pp_terminal.data.filters import filter_earlier_than, filter_by_type
from pp_terminal.utils.helper import enum_list_to_values
from .portfolio import Portfolio
from .schemas import TransactionType, TransactionSchema


_NEGATIVE_SECURITIES_ACCOUNT_TRANSACTION_TYPES = [
    TransactionType.SELL,
    TransactionType.DELIVERY_OUTBOUND,
    TransactionType.TRANSFER_OUT,
]

class PortfolioSnapshot:
    _per_date: datetime
    _portfolio: Portfolio

    def __init__(self, portfolio: Portfolio, per_date: datetime = datetime.now()):
        self._portfolio = portfolio
        self._per_date = per_date

    @property
    def date(self) -> datetime:
        return self._per_date

    @property
    def portfolio(self) -> Portfolio:
        return self._portfolio

    @property
    def prices(self) -> pd.DataFrame:
        return self._portfolio.prices.pipe(filter_earlier_than, target_date=self._per_date).sort_index(level='date', ascending=False)

    @property
    def latest_prices(self) -> pd.Series:
        prices = self.prices.groupby('securityId').head(1).reset_index('date')['price']
        prices.name = 'prices'

        return prices

    @property
    def securities_account_transactions(self) -> DataFrame[TransactionSchema]:
        transactions = self._portfolio.securities_account_transactions

        return TransactionSchema.validate(transactions.pipe(filter_earlier_than, target_date=self._per_date))

    @property
    def deposit_account_transactions(self) -> DataFrame[TransactionSchema]:
        transactions = self.portfolio.deposit_account_transactions

        return TransactionSchema.validate(transactions.pipe(filter_earlier_than, target_date=self._per_date))

    @property
    def shares(self) -> pd.Series | None:
        transactions = self.securities_account_transactions
        transactions['shares'] = transactions.apply(
            lambda row: -1 if row['type'] in enum_list_to_values(_NEGATIVE_SECURITIES_ACCOUNT_TRANSACTION_TYPES) else 1,
            axis=1
        ) * transactions['shares']

        shares = transactions.pipe(filter_by_type, transaction_types=[
            TransactionType.BUY,
            TransactionType.SELL,
            TransactionType.TRANSFER_IN,
            TransactionType.TRANSFER_OUT,
            TransactionType.DELIVERY_INBOUND,
            TransactionType.DELIVERY_OUTBOUND
        ]).groupby(['accountId', 'securityId', 'currency'])['shares'].sum()
        shares.name = 'shares'

        return shares

    @property
    def values(self) -> pd.Series:
        shares = self.shares
        if shares is None or shares.empty or self.latest_prices.empty:
            return pd.Series([], name='balance', index=pd.MultiIndex.from_tuples([], names=['accountId', 'securityId', 'currency']), dtype='float64')

        values = self.latest_prices * shares
        values.name = 'balance'

        return values.groupby(['accountId', 'securityId', 'currency']).sum()

    @property
    def balances(self) -> pd.Series:
        transactions = self.deposit_account_transactions

        balances = transactions.groupby(['accountId', 'currency'])['amount'].sum()
        balances.name = 'balance'

        return balances



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from abc import abstractmethod, ABC
from enum import Enum
from typing import Any

import pandas as pd
import rich
from rich.console import NewLine

from .table_decorator import TableDecorator, TableOptions


class OutputFormat(str, Enum):
    TABLE = 'table'
    CSV = 'csv'
    JSON = 'json'
    EXCEL = 'excel'


class Console(rich.console.Console):
    def print(self, *objects: Any, **kwargs: Any) -> None:
        kwargs['end'] = ''  # no newline at the end by default
        super().print(*objects, **kwargs)


class OutputStrategy(ABC):
    @abstractmethod
    def result_table(self, df: pd.DataFrame | None, options: TableOptions) -> Any:
        pass

    def hint(self, message: str) -> str:  # pylint: disable=unused-argument
        return ''

    def warning(self, message: str) -> str:  # pylint: disable=unused-argument
        return ''

    def empty_result(self) -> str:
        return ''

    def text(self, message: str) -> str:  # pylint: disable=unused-argument
        return ''


class RichOutputStrategy(OutputStrategy):
    def result_table(self, df: pd.DataFrame | None, options: TableOptions) -> Any:
        if df is None or df.empty:
            return self.empty_result(), NewLine()

        table = TableDecorator(options)
        table.add_df(df)

        return NewLine(), table, NewLine()

    def hint(self, message: str) -> str:
        return ':bulb: [bold]Hint:[/bold] ' + self.text(message)

    def warning(self, message: str) -> str:
        return ':backhand_index_pointing_right: [bold]Warning:[/bold] ' + self.text(message)

    def text(self, message: str) -> str:
        return message + "\n"

    def empty_result(self) -> str:
        return 'Nothing here..:sleeping: '


class CsvOutputStrategy(OutputStrategy):
    def result_table(self, df: pd.DataFrame | None, options: TableOptions) -> Any:
        if df is None:
            return self.empty_result()

        return (df.to_csv(index=options.show_index, float_format='%.2f'), )


class JsonOutputStrategy(OutputStrategy):
    def result_table(self, df: pd.DataFrame | None, options: TableOptions) -> Any:
        if df is None:
            return self.empty_result()

        return (df.to_json(index=options.show_index, orient='records'), )



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

from enum import Enum
from typing import Optional, TypeAlias, Any

import pandera.pandas as pa
from pandera.typing import DataFrame, Index, Series
from pydantic import BaseModel

Money: TypeAlias = float
Percent: TypeAlias = float


class TransactionType(Enum):
    BUY = "BUY"
    SELL = "SELL"
    DELIVERY_INBOUND = "DELIVERY_INBOUND"
    DELIVERY_OUTBOUND = "DELIVERY_OUTBOUND"
    TRANSFER_IN = "TRANSFER_IN"
    TRANSFER_OUT = "TRANSFER_OUT"
    DEPOSIT = "DEPOSIT"
    REMOVAL = "REMOVAL"
    INTEREST = "INTEREST"
    INTEREST_CHARGE = "INTEREST_CHARGE"
    FEES_REFUND = "FEES_REFUND"
    FEES = "FEES"
    DIVIDENDS = "DIVIDENDS"
    TAXES = "TAXES"
    TAX_REFUND = "TAX_REFUND"


class AccountType(Enum):
    SECURITIES = "portfolio"
    DEPOSIT = "account"


class TransactionSchema(pa.DataFrameModel):
    date: Index[pa.DateTime]
    accountId: Index[str]
    securityId: Index[str] = pa.Field(nullable=True)
    type: Series[str]  # @todo use pandera preprocessing?
    amount: Series[Money]
    shares: Series[float]
    accountType: Series[str]
    taxes: Series[Money] = pa.Field(default=0.0)
    fees: Optional[Series[Money]] = pa.Field(default=0.0, coerce=True)
    currency: Series[str] = pa.Field(nullable=True)


class Account(BaseModel):  # pylint: disable=too-few-public-methods
    accountId: str
    name: str
    type: str
    referenceAccount: Optional[str] = pa.Field(nullable=True)
    isRetired: Optional[bool] = pa.Field(coerce=True)
    currency: str | None
    additionalAttributes: dict[str, Any] = {}

class AccountSchema(pa.DataFrameModel):
    accountId: Index[str]
    name: Series[str]
    type: Series[str]  # @todo use pandera preprocessing?
    referenceAccount: Optional[Series[str]] = pa.Field(nullable=True)
    isRetired: Optional[Series[bool]] = pa.Field(coerce=True)
    currency: Series[str] = pa.Field(nullable=True)


class Security(BaseModel):  # pylint: disable=too-few-public-methods
    securityId: str
    name: str
    wkn: str | None
    currency: str | None
    isRetired: Optional[bool] = pa.Field(coerce=True)
    additionalAttributes: dict[str, Any] = {}

class SecuritySchema(pa.DataFrameModel):
    securityId: Index[str]
    name: Series[str]
    wkn: Series[str] = pa.Field(nullable=True)
    currency: Series[str] = pa.Field(nullable=True)
    isRetired: Optional[Series[bool]] = pa.Field(coerce=True)


class SecurityPriceSchema(pa.DataFrameModel):
    date: Index[pa.DateTime]
    securityId: Index[str]
    price: Series[Money]


class TaxPaidSchema(pa.DataFrameModel):
    year: Index[int] = pa.Field(coerce=True)
    account_id: Index[str]
    security_id: Index[str]
    tax_per_share: Series[Money]
    tax_free_allowance: Series[Money] = pa.Field(nullable=True, coerce=True)


class TaxLotSchema(pa.DataFrameModel):
    date: Index[pa.DateTime]  # Purchase date
    accountId: Index[str]
    securityId: Index[str]
    shares: Series[float]
    costBasis: Series[Money]
    purchasePrice: Series[Money] = pa.Field(nullable=True)
    currency: Series[str] = pa.Field(nullable=True)
    fees: Series[Money] = pa.Field(nullable=True)

    # Sell simulation (nullable until _calculate_sell_metrics)
    salePrice: Series[Money] = pa.Field(nullable=True)
    capitalGain: Series[Money] = pa.Field(nullable=True)
    grossProceeds: Series[Money] = pa.Field(nullable=True)
    prepaidTax: Series[Money] = pa.Field(nullable=True)
    taxableGain: Series[Money] = pa.Field(nullable=True, ge=0)
    totalTax: Series[Money] = pa.Field(nullable=True)
    netProceeds: Series[Money] = pa.Field(nullable=True)

    class Config:  # pylint: disable=too-few-public-methods
        add_missing_columns = True


class InterestResultSchema(pa.DataFrameModel):
    """Schema for interest calculation results."""
    accountId: Index[str]
    name: Series[str]
    currency: Series[str]
    mean_balance: Series[Money]
    interest: Series[Money]
    actual_interest: Series[Money] = pa.Field(nullable=True)

    @classmethod
    def empty(cls, *_args: Any) -> DataFrame['InterestResultSchema']:
        """Create empty DataFrame with correct index name.

        Overrides parent to fix Pandera limitation where .empty() doesn't preserve index names.
        """
        df = super().empty(*_args)
        df.index.name = 'accountId'
        return df


class VapResultSchema(pa.DataFrameModel):
    """Schema for Vorabpauschale (VAP) calculation results."""
    wkn: Series[str] = pa.Field(nullable=True, coerce=True)
    name: Series[str] = pa.Field(coerce=True)
    currency: Series[str] = pa.Field(coerce=True)

    class Config:  # pylint: disable=too-few-public-methods
        """Allow additional columns for dynamic account names."""
        strict = False  # Allow additional columns beyond those defined
        coerce = True



"""
    Copyright (C) 2025-26 Dipl.-Ing. Christoph Massmann <chris@dev-investor.de>

    This file is part of pp-terminal.

    pp-terminal is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    pp-terminal is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with pp-terminal. If not, see <http://www.gnu.org/licenses/>.
"""

import logging
from datetime import date
from typing import List, Any, Callable

import babel.numbers
import pandas as pd
import typer
from babel import Locale
from babel.numbers import format_currency
from typer.models import CommandFunctionType

from pp_terminal.domain.schemas import Money

log = logging.getLogger(__name__)

_PRECISION: int = 4


def set_precision(precision: int) -> None:
    global _PRECISION  # pylint: disable=global-statement
    _PRECISION = precision


def currency_exists(currency_code: str, locale: str | None = None) -> bool:
    return currency_code in Locale(str(locale)).currencies


def format_money(value: Money, currency: str | None = None, locale: str | None = babel.numbers.LC_NUMERIC) -> str:
    if pd.isna(value) or not isinstance(value, Money):
        return ''

    try:
        currency = currency if not currency is None and not pd.isna(currency) and currency_exists(currency, locale) else ''

        return format_currency(value, currency, locale=locale)
    except Exception:  # pylint: disable=broad-exception-caught
        # fallback e.g. if system locale is None/not set, or currency does not exist
        return f"{value:.2f}"


def format_shares(value: float) -> str:
    if pd.isna(value) or not isinstance(value, float):
        return ''

    return f"{float(value):.{_PRECISION}f}"


def enum_types_to_name(enum_list: List[Any]) -> List[Any]:
    # prepare for enum storage in dataframe
    for element in enum_list:
        element['type'] = element['type'].name

    return enum_list


def enum_list_to_values(enum_list: List[Any]) -> List[Any]:
    return [item.value for item in enum_list]


def run_all_group_cmds(app: typer.Typer) -> Callable[[CommandFunctionType], Callable[[typer.Context], CommandFunctionType]]:
    def decorator(func: CommandFunctionType) -> Callable[[typer.Context], CommandFunctionType]:
        def wrapper(ctx: typer.Context) -> Any:
            invoked_command = ctx.invoked_subcommand
            if ctx.invoked_subcommand is None:
                for command in app.registered_commands:
                    if command.callback is not None:
                        ctx.invoked_subcommand = command.name
                        log.debug('Running group command "%s"..', command.name)
                        command.callback(ctx)
            ctx.invoked_subcommand = invoked_command

            return func(ctx)
        return wrapper
    return decorator


def get_last_year() -> str:
    return str(date.today().year - 1)


def footer() -> str:
    return 'All results are non-binding and provided without any guarantee. Actual values may differ.'
